use heck::AsKebabCase;
use proc_macro::TokenStream;
use quote::{quote, ToTokens};
use syn::{parse_macro_input, Data, DeriveInput, Field, Ident};

pub fn arrow(item: TokenStream) -> TokenStream {
    let input = parse_macro_input!(item as DeriveInput);
    let Data::Struct(structure) = input.data else {
        panic!("Only structs are supported as arrows")
    };
    match structure.fields {
        syn::Fields::Named(fields) => from_named(&input.ident, fields),
        syn::Fields::Unnamed(f) if f.unnamed.len() == 1 => {
            let first = f.unnamed.first().unwrap();
            from_newtype(&input.ident, first)
        }
        _ => panic!(
            "Only newtype structs and structs with named fields can have a derived arrow impl"
        ),
    }
}

fn from_named(name: &Ident, fields: syn::FieldsNamed) -> TokenStream {
    let (origin, target, identity) = extract_idents(fields);
    match identity {
        Some(id) => make_multi_arrow(name, origin, target, id),
        None => make_basic_arrow(name, origin, target),
    }
}

fn make_basic_arrow(name: &Ident, origin: Ident, target: Ident) -> TokenStream {
    let spec = gen_spec(name);
    TokenStream::from(quote! {
        #spec
        impl store::arrow::Arrow for #name {}
        impl From<store::arrow::Basic> for #name {
            fn from(v: store::arrow::Basic) -> #name {
                #name {
                    #origin: v.origin,
                    #target: v.target,
                }
            }
        }
        impl From<#name> for store::arrow::Basic {
            fn from(v: #name) -> store::arrow::Basic {
                store::arrow::Basic {
                    origin: v.#origin,
                    target: v.#target,
                }
            }
        }
    })
}

fn make_multi_arrow(name: &Ident, origin: Ident, target: Ident, id: Ident) -> TokenStream {
    let spec = gen_spec(name);
    TokenStream::from(quote! {
        #spec
        impl store::arrow::Arrow for #name {
            type Kind = store::arrow::Multi;
        }
        impl From<store::arrow::Multi> for #name {
            fn from(v: store::arrow::Multi) -> #name {
                #name {
                    #id: v.identity,
                    #origin: v.origin,
                    #target: v.target,
                }
            }
        }
        impl From<#name> for store::arrow::Multi {
            fn from(v: #name) -> store::arrow::Multi {
                store::arrow::Multi {
                    identity: v.#id,
                    origin: v.#origin,
                    target: v.#target,
                }
            }
        }
    })
}

fn extract_idents(fields: syn::FieldsNamed) -> (Ident, Ident, Option<Ident>) {
    let origin = extract_ident("origin", &fields).unwrap();
    let target = extract_ident("target", &fields).unwrap();
    let id = extract_ident("identity", &fields);
    (origin, target, id)
}

fn extract_ident(name: &str, fields: &syn::FieldsNamed) -> Option<Ident> {
    // Prefer marked fields and default to correctly named fields.
    fields
        .named
        .iter()
        .find(|field| {
            field
                .attrs
                .iter()
                .filter_map(|attr| attr.meta.path().get_ident())
                .any(|id| id == name)
        })
        .and_then(|f| f.ident.clone())
        .or_else(|| {
            fields
                .named
                .iter()
                .filter_map(|f| f.ident.clone())
                .find(|id| id == name)
        })
}

fn gen_spec(name: &Ident) -> impl ToTokens {
    let prefix = AsKebabCase(name.to_string());
    let by_origin = format!("{prefix}/by-origin");
    let by_target = format!("{prefix}/by-target");
    quote! {
        impl store::types::DataType for #name {
            type Type = store::types::ArrowSpec;
            const SPEC: Self::Type = store::types::ArrowSpec {
                by_origin: store::types::Keyspace(#by_origin),
                by_target: store::types::Keyspace(#by_target),
            };
        }
    }
}

fn from_newtype(name: &Ident, field: &Field) -> TokenStream {
    let spec = gen_spec(name);
    let typ = &field.ty;
    TokenStream::from(quote! {
        #spec
        impl store::arrow::Arrow for #name {
            type Kind = #typ;
        }
        impl From<#typ> for #name {
            fn from(v: #typ) -> #name { #name(v) }
        }
        impl From<#name> for #typ {
            fn from(v: #name) -> #typ { v.0 }
        }
    })
}
