use fetch::Client;
use store::{Key, Store, Transaction};

use crate::{config::Config, Result};

/// The context of a running ActivityPuppy.
///
/// This type provides access to the data store and configuration.
#[derive(Clone)]
pub struct Context {
    config: Config,
    client: Client,
    store: Store,
}

impl Context {
    /// Load the server context from the configuration.
    pub fn load(config: Config) -> Result<Context> {
        let store = Store::open(&config.state_dir, crate::data::schema())?;
        let client = Client::new();
        Ok(Context { config, store, client })
    }
    /// Do a data store [transaction][store::Transaction].
    pub fn run<T>(&self, f: impl FnOnce(&Transaction<'_>) -> Result<T>) -> Result<T> {
        self.store.run(f)
    }
    /// Access the store directly.
    pub fn store(&self) -> &Store {
        &self.store
    }
    /// Access the configuration.
    pub fn config(&self) -> &Config {
        &self.config
    }
    /// Create an ActivityPub object ID from a key.
    pub fn mk_url(&self, key: Key) -> String {
        format!("https://{}/o/{key}", self.config.ap_domain)
    }
    /// Access the federation client.
    pub fn resolver(&self) -> &Client {
        &self.client
    }
}

/// Load a context for running tests in.
#[cfg(test)]
pub fn test_context<T>(
    config: Config,
    schema: store::types::Schema,
    test: impl FnOnce(Context) -> Result<T>,
) -> Result<T> {
    let client = Client::new();
    Store::test(schema, |store| test(Context { config, store, client }))
}
