//! Datas for the data store!
//!
//! This module contains the definitions for the data store.

use bincode::{Decode, Encode};
use derive_more::Display;
use store::{types::Schema, Alias, Arrow, Key, Mixin};

/// *Bites you*
#[derive(Arrow, Debug, PartialEq, Eq, Clone, Copy)]
pub struct Bite {
    #[identity]
    pub id: Key,
    #[origin]
    pub biter: Key,
    #[target]
    pub victim: Key,
}

/// Properties of ActivityPub objects.
#[derive(Mixin, Encode, Decode, Debug, Clone)]
pub struct Object {
    /// The ActivityPub id of the object.
    #[index]
    pub id: Id,
    /// What kind of object is it?
    pub kind: ObjectKind,
    /// Whether or not the object resides on this server or on another one.
    pub local: bool,
}

/// Allows case analysis on the type of ActivityPub objects.
#[derive(Encode, Decode, Debug, Clone)]
pub enum ObjectKind {
    Actor,
    Activity(ActivityKind),
    Notelike(String),
}

/// The type of an activity.
#[derive(Encode, Decode, Debug, Clone)]
pub enum ActivityKind {
    /// Used for posting stuff!
    Create = 0,
    /// Represents a follow request.
    Follow = 1,
    /// Used to signal that a follow request was accepted.
    Accept = 2,
    /// Used to reject a follow request.
    Reject = 3,
    /// See [`bites`](crate::bites).
    Bite = 4,
}

#[derive(Mixin, Encode, Decode, Debug, Clone)]
pub struct Channel {
    pub inbox: String,
    // TODO: add public key here
}

/// A predicate; `follower` "follows" `followed`.
#[derive(Arrow, Debug, PartialEq, Eq, Clone, Copy)]
pub struct Follows {
    #[origin]
    pub follower: Key,
    #[target]
    pub followed: Key,
}

/// An instance of a request from some `origin` user to follow a `target` user.
///
/// This should not be used to determine whether two actors are following each other. For that, use
/// [`Follows`], a basic arrow for exactly this purpose. *This* arrow is used to identify specific
/// instances of *requests*, and serves mostly as a historical reference and for synchronizing with
/// other servers.
///
/// Used to represent a `Follow` activity.
///
/// Mixins always present for the `id`:
///
/// - [`Status`], carrying the status of the request.
#[derive(Arrow, Debug, PartialEq, Eq, Clone, Copy)]
pub struct FollowRequest {
    /// The unique ID of this particular request.
    #[identity]
    pub id: Key,
    /// The "follower", the user that made the request.
    pub origin: Key,
    /// The one the request is made to.
    pub target: Key,
}

/// The status of a [`FollowRequest`].
///
/// Valid state transitions:
///
/// ```text
///                      ┌──────────────▶  Rejected
///                      │                         
///                      │                         
///                      │                         
///                                                
/// None  ─────────▶  Pending   ────────▶  Accepted
///                                                
///                      │                    │    
///                      │                    │    
///                      │                    │    
///                      ▼                    │    
///                  Withdrawn   ◀────────────┘        
/// ```
///
/// In addition, a follow request will be deleted if either endpoint is removed from the graph.
#[derive(Mixin, Encode, Decode, Eq, PartialEq, Clone)]
pub enum Status {
    /// The follow request was previously pending or accepted, but since withdrawn.
    ///
    /// This can happen when someone cancels their follow request or unfollows the target.
    Withdrawn,
    /// The follow request was accepted.
    Accepted,
    /// The follow request was denied.
    Rejected,
    /// The follow request is still under review.
    Pending,
}

/// An ActivityPub ID, used to look up remote objects by their canonical URL.
#[derive(Alias, Encode, Decode, Clone, PartialEq, Eq, Debug, Hash, Display)]
pub struct Id(pub String);

/// A "profile" in the social media sense.
///
/// Contains all presentation information about someone making posts.
#[derive(Mixin, Encode, Decode, Debug, Clone)]
pub struct Profile {
    /// How many posts has this user made?
    pub post_count: usize,
    /// The name used for the profile's handle.
    #[index] // <- currently doesnt do anything but i have an idea
    pub account_name: Username,
    /// The name displayed above their posts.
    pub display_name: Option<String>,
    /// The "bio", a freeform "about me" field.
    pub about_string: Option<String>,
    /// Arbitrary custom metadata fields.
    pub about_fields: Vec<(String, String)>,
}

/// A unique name for an actor that is part of their "handle".
#[derive(Alias, Encode, Decode, Clone, PartialEq, Eq, Debug, Hash, Display)]
pub struct Username(pub String);

/// The relation that `author` has constructed and published `object`.
#[derive(Arrow, Debug, PartialEq, Eq, Clone, Copy)]
pub struct AuthorOf {
    #[origin]
    pub author: Key,
    #[target]
    pub object: Key,
}

/// The contents of a post.
#[derive(Mixin, Encode, Decode, Debug, Clone, Default)]
pub struct Content {
    /// Main post body.
    pub content: Option<String>,
    /// Content warning for the post.
    pub warning: Option<String>,
}

/// A public key used for verifying requests.
#[derive(Mixin, Encode, Decode, Debug, Clone)]
pub struct PublicKey {
    pub key_id: String,
    pub key_pem: String,
}

/// A private key for signing requests with.
#[derive(Mixin, Encode, Decode, Clone)]
pub struct PrivateKey {
    pub key_pem: String,
}

/// Represents a `Create` activity.
#[derive(Arrow)]
pub struct Create {
    #[identity]
    pub id: Key,
    #[origin]
    pub actor: Key,
    #[target]
    pub object: Key,
}

/// Construct the schema.
pub fn schema() -> Schema {
    Schema::new()
        // Mixins
        .has::<Profile>()
        .has::<Content>()
        .has::<Status>()
        .has::<Object>()
        .has::<Channel>()
        .has::<PrivateKey>()
        .has::<PublicKey>()
        // Aliases
        .has::<Username>()
        .has::<Id>()
        // Arrows
        .has::<Bite>()
        .has::<FollowRequest>()
        .has::<AuthorOf>()
        .has::<Follows>()
        .has::<Create>()
}
