//! Interactions between actors.

use store::{util::IterExt as _, Key, StoreError, Transaction};

use crate::{
    actor::Actor,
    data::{FollowRequest, Bite, Status, Follows},
    Context, Error, Result,
};

/// Interactions with other objects.
impl Actor {
    /// Create a [`Bite`].
    pub fn bite(&self, victim: Actor) -> Bite {
        Bite {
            victim: victim.key,
            biter: self.key,
            id: Key::gen(),
        }
    }
    /// Construct a [`FollowRequest`].
    pub fn follow_request(&self, target: Actor) -> FollowRequest {
        FollowRequest {
            origin: self.key,
            target: target.key,
            id: Key::gen(),
        }
    }
    /// Makes `biter` bite `victim` and inserts the records into the database.
    pub fn do_bite(&self, cx: &Context, victim: Actor) -> Result<Bite> {
        let bite = self.bite(victim);
        cx.run(|tx| tx.create(bite).map_err(Error::Store))?;
        Ok(bite)
    }
    /// Creates a follow request from `self` to `target`.
    pub fn do_follow_request(&self, cx: &Context, target: Actor) -> Result<FollowRequest> {
        let req = self.follow_request(target);
        cx.run(|tx| {
            tx.create(req)?;
            tx.add_mixin(req.id, Status::Pending)?;
            Ok(req)
        })
    }
    /// Accept a follow request.
    pub fn do_accept_request(&self, cx: &Context, req: FollowRequest) -> Result<()> {
        debug_assert! {
            self.key == req.target,
            "only the target of a follow request may accept it"
        };
        cx.run(|tx| try {
            let fr = tx
                .between::<FollowRequest>(req.origin, req.target)
                // Get the one that is equal to `req`.
                .filter(|fr| fr.as_ref().is_ok_and(|f| f.id == req.id))
                .last()
                .unwrap()?;
            // Only apply the update if the follow request is still in a pending state.
            if let Some(Status::Pending) = tx.get_mixin(fr.id)? {
                tx.update(fr.id, |_| Status::Accepted)?;
                tx.create(Follows {
                    follower: req.origin,
                    followed: req.target,
                })?;
            }
        })
    }
    /// Reject a follow request.
    pub fn do_reject_request(&self, cx: &Context, req: FollowRequest) -> Result<()> {
        debug_assert! {
            self.key == req.target,
            "only the target of a follow request may accept it"
        };
        cx.run(|tx| try { tx.update(req.id, |_| Status::Rejected)? })?;
        Ok(())
    }
    /// Get all pending follow request for `self`.
    pub fn pending_requests<'c>(
        &self,
        tx: &'c Transaction<'c>,
    ) -> impl Iterator<Item = Result<FollowRequest>> + 'c {
        tx.incoming::<FollowRequest>(self.key)
            .map_err(Error::Store)
            .filter_bind_results(|req| Ok(if req.is_pending(tx)? { Some(req) } else { None }))
    }
    /// Get all nodes `self` is following.
    pub fn following<'c>(&self, tx: &'c Transaction<'c>) -> impl Iterator<Item = Result<Key>> + 'c {
        tx.outgoing::<Follows>(self.key)
            .map_err(Error::Store)
            .map_ok(|a| a.followed)
    }
    /// Get all followers of `self`.
    pub fn followers<'c>(&self, tx: &'c Transaction<'c>) -> impl Iterator<Item = Result<Key>> + 'c {
        tx.incoming::<Follows>(self.key)
            .map_err(Error::Store)
            .map_ok(|a| a.follower)
    }
    /// List all specific times `self` was bitten.
    pub fn bites_suffered<'c>(
        &self,
        tx: &'c Transaction<'c>,
    ) -> impl Iterator<Item = Result<Bite>> + 'c {
        tx.incoming::<Bite>(self.key).map_err(Error::Store)
    }
    /// Check whether `self` follows `other`.
    pub fn follows(&self, tx: &Transaction<'_>, other: &Actor) -> Result<bool> {
        try { tx.exists::<Follows>(self.key, other.key)? }
    }
}

impl FollowRequest {
    /// Determine if this follow request is pending.
    pub fn is_pending(&self, tx: &Transaction<'_>) -> Result<bool> {
        // The status is stored as a mixin, so we need to get it.
        let Some(st) = tx.get_mixin::<Status>(self.id)? else {
            // If we don't have a status for a follow request, something is borked.
            return Err(StoreError::Missing.into());
        };
        // If the status of the follow request is pending, it can't also be true that the follows
        // relation already exists.
        debug_assert! {
            !(st == Status::Pending)
                || tx.exists::<Follows>(self.origin, self.target).map(|x| !x)?,
            "fr.is_pending -> !(fr.origin follows fr.target)"
        };
        Ok(st == Status::Pending)
    }
}

#[cfg(test)]
mod tests {
    use store::util::IterExt as _;

    use crate::{
        actor::{create_local as create_actor, Actor},
        config::Config,
        data::{schema, FollowRequest, Follows},
        test_context, Context, Result,
    };

    fn make_test_actors(cx: &Context) -> Result<(Actor, Actor)> {
        let alice = create_actor(&cx, "alice")?;
        let bob = create_actor(&cx, "bob")?;
        eprintln!("alice={alice:?}, bob={bob:?}");
        Ok((alice, bob))
    }
    fn test_config() -> Config {
        Config {
            ap_domain: String::from("unit-test.puppy.gay"),
            wf_domain: String::from("unit-test.puppy.gay"),
            state_dir: todo!(), // TODO: make this a temp dir
            port: 0,
        }
    }

    #[test]
    fn create_fr() -> Result<()> {
        test_context(test_config(), schema(), |cx| {
            let (alice, bob) = make_test_actors(&cx)?;
            alice.do_follow_request(&cx, bob)?;
            assert!(
                cx.store().exists::<FollowRequest>(alice.key, bob.key)?,
                "(alice -> bob) ∈ follow-requested"
            );
            assert!(
                !cx.store().exists::<Follows>(alice.key, bob.key)?,
                "(alice -> bob) ∉ follows"
            );
            let pending_for_bob = cx.run(|tx| {
                bob.pending_requests(&tx)
                    .map_ok(|fr| fr.origin)
                    .try_collect::<Vec<_>>()
            })?;
            assert_eq!(pending_for_bob, vec![alice.key], "bob.pending = {{alice}}");
            Ok(())
        })
    }

    #[test]
    fn accept_fr() -> Result<()> {
        test_context(test_config(), schema(), |cx| {
            let db = cx.store();
            let (alice, bob) = make_test_actors(&cx)?;
            let req = alice.do_follow_request(&cx, bob)?;
            bob.do_accept_request(&cx, req)?;

            assert!(
                db.exists::<Follows>(alice.key, bob.key)?,
                "(alice -> bob) ∈ follows"
            );
            assert!(
                !db.exists::<Follows>(bob.key, alice.key)?,
                "(bob -> alice) ∉ follows"
            );

            cx.run(|tx| try {
                let pending_for_bob: Vec<_> = bob.pending_requests(&tx).try_collect()?;
                assert!(pending_for_bob.is_empty(), "bob.pending = ∅");

                let followers_of_bob: Vec<_> = bob.followers(&tx).try_collect()?;
                assert_eq!(
                    followers_of_bob,
                    vec![alice.key],
                    "bob.followers = {{alice}}"
                );
            })
        })
    }

    #[test]
    fn listing_follow_relations() -> Result<()> {
        test_context(test_config(), schema(), |cx| try {
            let (alice, bob) = make_test_actors(&cx)?;
            let req = alice.do_follow_request(&cx, bob)?;
            bob.do_accept_request(&cx, req)?;

            cx.run(|tx| try {
                let followers_of_bob: Vec<_> = bob.followers(&tx).try_collect()?;
                assert_eq!(
                    followers_of_bob,
                    vec![alice.key],
                    "bob.followers = {{alice}}"
                );

                let following_of_alice: Vec<_> = alice.following(&tx).try_collect()?;
                assert_eq!(
                    following_of_alice,
                    vec![bob.key],
                    "alice.following = {{bob}}"
                );
            })?
        })
    }
}
