//! Timelines: where you go to view the posts.

use std::ops::RangeBounds;

use chrono::{DateTime, Utc};
use either::Either::{Left, Right};
use fetch::object::{Activity, Note, Object};
use store::{util::IterExt as _, Key, Store, StoreError, Transaction};

use crate::{
    actor::{get_signing_key, Actor},
    data::{
        self, ActivityKind, AuthorOf, Channel, Content, Create, Follows, Id, ObjectKind, Profile,
    },
    Context,
};

#[derive(Clone, Copy, Debug)]
pub struct Post {
    pub key: Key,
}

impl From<&str> for Content {
    fn from(value: &str) -> Self {
        value.to_string().into()
    }
}

impl From<String> for Content {
    fn from(value: String) -> Self {
        Content {
            content: Some(value),
            warning: None,
        }
    }
}

/// A piece of content posted by someone.
#[derive(Clone, Debug)]
pub struct PostData {
    /// The post's internal ID.
    pub id: Post,
    /// The actual post contents.
    pub content: Content,
    /// Metadata about the post's author.
    pub author: Author,
}

/// Information about a [`Post`]'s author.
#[derive(Clone, Debug)]
pub struct Author {
    /// The identifier of the author.
    pub id: Key,
    /// The name to display along with the post.
    pub display_name: String,
    /// An informal identifier for a particular author.
    pub handle: String,
}

/// An ordered list of [`Post`]s for viewing.
#[derive(Debug)]
pub struct Timeline {
    items: Vec<Item>,
}

/// Discrete events that can be displayed to a user as part of a timeline.
#[derive(Debug)]
enum Item {
    Post(PostData),
}

impl Item {
    /// Get the timeline item if it is a [`Post`].
    pub fn as_post(&self) -> Option<&PostData> {
        match self {
            Item::Post(ref post) => Some(post),
        }
    }
}

impl Timeline {
    /// Get all the posts in the timeline.
    pub fn posts(&self) -> impl Iterator<Item = &PostData> {
        self.items.iter().filter_map(|x| x.as_post())
    }
}

/// Gets at most `limit` of the posts known to the instance that were inserted within `time_range`.
pub fn fetch_timeline(
    db: &Store,
    time_range: impl RangeBounds<DateTime<Utc>>,
    limit: Option<usize>,
) -> Result<Timeline, StoreError> {
    let posts = db.run(|tx| {
        // Get all post content entries (the argument passed here is a range of chrono datetimes).
        let iter = tx.range::<Content>(time_range);
        let iter = match limit {
            Some(n) => Left(iter.take(n)),
            None => Right(iter),
        };
        // Then, we're gonna map each of them to their author, and get the profile information needed to
        // render the post (mostly display name and handle).
        iter.bind_results(|(key, content)| try {
            // Take the first author. There is nothing stopping a post from having multiple authors, but
            // let's take it one step at a time.
            let (author, Some(Profile { display_name, account_name, .. })) = tx
                .join_on(|a: AuthorOf| a.author, tx.incoming(key))?
                .swap_remove(0)
            else {
                // We expect all posts to have at least one author, so we should complain if there is one
                // that doesn't (for now). For robustness, the `.collect()` down there should be replaced
                // with a strategy where we log a warning instead of failing, but in the current state of
                // the project, failing fast is a good thing.
                return Err(StoreError::Missing);
            };
            Item::Post(PostData {
                id: Post { key },
                author: Author {
                    id: author,
                    handle: format!("@{account_name}"),
                    display_name: display_name.unwrap_or(account_name.0),
                },
                content,
            })
        })
        .collect()
    })?;
    Ok(Timeline { items: posts })
}

/// Create a new post entity.
pub fn create_local_post(
    cx: &Context,
    author: Key,
    content: impl Into<Content>,
) -> crate::Result<Post> {
    let content = content.into();
    cx.run(|tx| {
        let key = Key::gen();
        // Local stuff
        mixin_post(tx, key, author, content)?;
        // Federation stuff
        let id = Id(cx.mk_url(key));
        tx.add_alias(key, id.clone())?;
        tx.add_mixin(key, data::Object {
            kind: ObjectKind::Notelike("Note".to_string()),
            local: true,
            id,
        })?;
        Ok(Post { key })
    })
}

/// Assumes all objects referenced already exist.
#[tracing::instrument(skip(cx))]
pub fn create_post_from_note(cx: &Context, note: Note) -> crate::Result<Post> {
    cx.run(|tx| {
        let Some(author) = tx.lookup(Id(note.author))? else {
            panic!("needed author to already exist")
        };

        let key = Key::gen();

        tx.add_alias(key, Id(note.id.clone()))?;
        tx.create(AuthorOf { object: key, author })?;
        tx.add_mixin(key, Content {
            content: note.content,
            warning: note.summary,
        })?;
        tx.add_mixin(key, data::Object {
            kind: ObjectKind::Notelike(note.kind),
            id: Id(note.id),
            local: false,
        })?;

        Ok(Post { key })
    })
}

#[tracing::instrument(skip(cx))]
pub async fn federate_post(cx: &Context, post: Post) -> crate::Result<()> {
    // Obtain all the data we need to construct our activity
    let (Content { content, warning }, url, author, signing_key, followers) = cx.run(|tx| try {
        let Some(AuthorOf { author, .. }) = tx.incoming(post.key).next().transpose()? else {
            panic!("can't federate post without author: {post:?}")
        };
        let signing_key = get_signing_key(tx, Actor { key: author })?;
        let (c, data::Object { id, .. }) = tx.get_mixin_many(post.key)?;
        let targets = tx.join_on::<Channel, _>(|a| a.follower, tx.incoming::<Follows>(author))?;
        (c, id, author, signing_key, targets)
    })?;

    let activity_key = Key::gen();

    // Insert a create activity into the database so we can serve it later
    cx.run(|tx| try {
        let id = Id(cx.mk_url(activity_key));
        tx.add_alias(activity_key, id.clone())?;
        tx.add_mixin(activity_key, data::Object {
            kind: ObjectKind::Activity(ActivityKind::Create),
            local: true,
            id,
        })?;
        tx.create(Create {
            id: activity_key,
            actor: author,
            object: post.key,
        })?;
    })?;

    // Construct an ActivityPub message to send
    let activity = Activity {
        id: cx.mk_url(activity_key),
        actor: signing_key.owner.clone(),
        object: Box::new(Object::Note(Note {
            id: url.to_string(),
            kind: "Note".to_string(),
            author: cx.mk_url(author),
            summary: warning,
            content,
        })),
        kind: "Create".to_string(),
    };

    for inbox in followers
        .into_iter()
        .filter_map(|(_, c)| c.map(|t| t.inbox))
        // FIXME: remove this when im done testing
        .chain(["https://crimew.gay/users/riley/inbox".to_string()])
    {
        fetch::deliver(&signing_key, &activity, &inbox).await;
    }

    Ok(())
}

/// Add a post's mixins and predicates to an existing `node`.
pub fn mixin_post(
    tx: &Transaction<'_>,
    node: Key,
    author: Key,
    content: impl Into<Content>,
) -> Result<Key, StoreError> {
    tx.update::<Profile>(author, |mut profile| {
        profile.post_count += 1;
        profile
    })?;
    tx.add_mixin(node, content.into())?;
    tx.create(AuthorOf { author, object: node })?;
    Ok(node)
}

pub fn list_posts_by_author(db: &Store, author: Key) -> Result<Vec<(Key, Content)>, StoreError> {
    db.run(|tx| {
        let posts = tx
            .join_on(|a: AuthorOf| a.object, tx.outgoing(author))?
            .into_iter()
            .filter_map(|(k, opt)| try { (k, opt?) })
            .collect();
        Ok(posts)
    })
}
