/// Derive an implementation of [`Alias`].
pub use r#macro::Alias;

use super::{
    types::{AliasSpec, DataType},
    Batch, Store, Transaction,
};
use crate::{Key, Result};

/// An alternative unique identifier for a node.
pub trait Alias: DataType<Type = AliasSpec> + From<String> + AsRef<str> {}

impl Transaction<'_> {
    /// Look up the key associated with the alias.
    pub fn lookup<A>(&self, alias: A) -> Result<Option<Key>>
    where
        A: Alias,
    {
        op::lookup::<A>(self, alias.as_ref())
    }
    /// Get the alias associated with the `node`.
    pub fn get_alias<A>(&self, node: Key) -> Result<Option<A>>
    where
        A: Alias,
    {
        op::get_alias(self, node)
    }
    /// Add an alias to `node`.
    pub fn add_alias<A>(&self, node: Key, alias: A) -> Result<()>
    where
        A: Alias,
    {
        op::add_alias::<A>(self, node, alias.as_ref())
    }
    /// Check whether `node` has an `M` defined for it.
    pub fn has_alias<A>(&self, node: Key) -> Result<bool>
    where
        A: Alias,
    {
        op::has_alias::<A>(self, node)
    }
}

impl Store {
    /// Look up the key associated with the alias.
    pub fn lookup<A>(&self, alias: A) -> Result<Option<Key>>
    where
        A: Alias,
    {
        op::lookup::<A>(self, alias.as_ref())
    }
    /// Get the alias associated with the `node`.
    pub fn get_alias<A>(&self, node: Key) -> Result<Option<A>>
    where
        A: Alias,
    {
        op::get_alias(self, node)
    }
    /// Check whether `node` has an `M` defined for it.
    pub fn has_alias<A>(&self, node: Key) -> Result<bool>
    where
        A: Alias,
    {
        op::has_alias::<A>(self, node)
    }
}

impl Batch {
    /// Add an alias to `node`.
    ///
    /// # Warning
    ///
    /// This will *not* fail if the key already has a alias of this type, and in fact *it will cause fundamental inconsistency*
    /// if the alias already exists. Don't call this function unless you know that neither `node` nor `alias` exist yet.
    pub fn put_alias<A>(&mut self, node: Key, alias: A)
    where
        A: Alias,
    {
        // TODO: consistency *could* be checked by manually iterating over the transaction using `WriteBatch::iterate`
        op::add_alias::<A>(self, node, alias.as_ref()).unwrap();
    }
}

mod op {
    use crate::{internal::*, Alias, Key, Result, OK};

    pub fn lookup<A: Alias>(cx: &impl Query, alias: &str) -> Result<Option<Key>> {
        cx.open(A::SPEC.keyspace).get(alias).map(|k| match k {
            Some(x) => Some(Key::from_slice(x.as_ref())),
            None => None,
        })
    }

    pub fn has_alias<A: Alias>(cx: &impl Query, node: Key) -> Result<bool> {
        cx.open(A::SPEC.reversed).has(node)
    }

    pub fn add_alias<A: Alias>(cx: &impl Write, node: Key, alias: &str) -> Result<()> {
        cx.open(A::SPEC.keyspace).set(alias, node)?;
        cx.open(A::SPEC.reversed).set(node, alias)?;
        OK
    }

    pub fn get_alias<A: Alias>(cx: &impl Query, node: Key) -> Result<Option<A>> {
        let buf = cx.open(A::SPEC.reversed).get(node)?;
        Ok(buf.map(decode))
    }

    fn decode<T>(data: impl AsRef<[u8]>) -> T
    where
        T: From<String>,
    {
        T::from(String::from_utf8_lossy(data.as_ref()).into_owned())
    }
}
