//! Directed edges, both parallel and simple.
//!
//! This module's main exports are [`Arrow`], and the two kinds of arrows: [`Basic`] and [`Multi`].
//!
//! Querying information about arrows can be done using the APIs exposed by [`Store`] and [`Transaction`],
//! and manipulating them can likewise be done from within the context of a `Transaction`.
//!
//! The arrow API is designed to aggressively minimize boilerplate for defining arrow types, and uses a
//! few tricks to do with associated constants and types to make it all work nicely.
//!
//! # Terminology
//!
//! An arrow is a part of a graph. Graphs consist of *nodes* (also called *vertices*) and *edges*. Nodes
//! can be seen as "things", and edges as connections between those things, defined by the two nodes that
//! they connect (which are called the *endpoints* of the edge).
//!
//! These edges can be directed or undirected. The difference is that undirected edges are identified by
//! an unordered pair of their endpoints, whereas directed edges (also called **arrows**), are identified
//! by an ordered pair, where one of the endpoints is the *tail* (or *origin* in the code/docs here) and
//! the other is the *head* (usually called *target* here).
//!
//! # Arrow kinds
//!
//! Arrows can be either [`Basic`] or [`Multi`]. The main difference is that basic arrows are defined
//! solely by which two nodes they connect, which means that their representation and certain operations
//! are more efficient. The trade-off is that they cannot capture more complex information than "this
//! edge exists".
//!
//! For some use cases (for example, predicates) this is sufficient, but other use cases require multiple,
//! individually identifiable and manipulatable parallel edges. Here, the trade-off is that while they
//! are much more expressive, and can be labeled by associating [mixins] with the arrow's identity key,
//! they incur more space overhead, and most operations on them are more expensive compared to basic
//! edges.
//!
//! Most arrow operations work on either kind of edge. Some signatures reference [`Arrow::Kind`], which
//! is either of the `Multi` or `Basic` types mentioned before. Because parallel arrows need to be
//! discernable from each other, each of them also has an `identity` key, in addition to listing the two
//! edges they connect.
//!
//! [mixins]: super::Mixin
#![allow(private_interfaces)]

pub use self::kinds::{Basic, Multi};
use super::{
    types::{ArrowSpec, DataType},
    Batch, Store, Transaction,
};
use crate::{internal::Context as _, util::IterExt as _, Key, Result};

/// A directed edge.
///
/// See the [module docs][self] for an introduction.
pub trait Arrow: DataType<Type = ArrowSpec> + From<Self::Kind> + Into<Self::Kind> {
    /// The representation of this arrow, which also determines whether parallel edges are allowed.
    type Kind: ArrowKind = Basic;
}

/// Parameterizing arrows so we can distinguish between kinds of arrows.
///
/// This lets us present a common API for certain arrow-related operations while also leveraging some
/// specialization. Essentially, from a type parameter which implements [`Arrow`], we can tell both at
/// the type level and at the value level whether that arrow is a multi-arrow or not.
pub trait ArrowKind {
    /// Whether this kind of arrow should be represented using the specialized representation for edges
    /// that are allowed to be parallel.
    const IS_MULTI: bool;
    /// Construct an arrow from a buffer containing a correctly-oriented arrow.
    ///
    /// Each arrow is stored twice, once "correctly", and once "reversed". This allows us to efficiently
    /// list both the outgoing and incoming edges for any particular vertex by using a prefix scan on the
    /// [`BY_ORIGIN`][ArrowSpec::by_origin] and [`BY_TARGET`][ArrowSpec::by_target] keyspaces respectively.
    ///
    /// The buffer passed to this function will start with 16 bytes origin, followed by 16 bytes target.
    /// For basic arrows, that's it, but for multiarrows there is an additional 16 bytes of "identity",
    /// which is needed to discriminate between multiple parallel edges.
    ///
    /// # Failure
    ///
    /// This method must panic if `buf` is not the expected size (32 bytes for basic arrows, 48 bytes for
    /// multi arrows). The responsibility for ensuring that `buf` is correctly oriented lies with the
    /// caller lest the result is incorrect, but passing an incorrectly oriented arrow is not a memory
    /// safety issue, so this function is safe.
    fn dec(buf: &[u8]) -> Self;
    /// Encode an arrow's key origin-first and target-first.
    fn enc(self) -> (impl AsRef<[u8]>, impl AsRef<[u8]>);
    #[doc(hidden)]
    /// Turn him into a raw edge.
    fn raw(&self) -> Raw;
}

union Raw {
    multi: Multi,
    basic: Basic,
}

impl Store {
    /// Check whether there exists any arrow of type `A` that points from `origin` to `target`.
    pub fn exists<A>(&self, origin: Key, target: Key) -> Result<bool>
    where
        A: Arrow,
    {
        op::exists::<A>(self, origin, target)
    }
    /// Get all arrows of type `A` that point at `target`.
    pub fn incoming<'a, A>(&'a self, target: Key) -> impl Iterator<Item = Result<A>> + 'a
    where
        A: Arrow + 'a,
    {
        op::incoming::<A>(self, target).map_ok(A::from)
    }
    /// Get all arrows of type `A` that point away from `origin`.
    pub fn outgoing<'a, A>(&'a self, origin: Key) -> impl Iterator<Item = Result<A>> + 'a
    where
        A: Arrow + 'a,
    {
        op::outgoing::<A>(self, origin).map_ok(A::from)
    }
    /// List all arrows between `a` and `b`, in either direction.
    pub fn between<'a, A>(&'a self, a: Key, b: Key) -> impl Iterator<Item = Result<A>> + 'a
    where
        A: Arrow<Kind = Multi> + 'a,
    {
        op::between::<A>(self, a, b).map_ok(A::from)
    }
    /// Construct the arrow from its identifier.
    pub fn get_arrow<A>(&self, key: Key) -> Result<Option<A>>
    where
        A: Arrow<Kind = Multi>,
    {
        let arrow = self
            .open(crate::types::MULTIEDGE_HEADERS)
            .get(key)?
            .map(|v| Key::split(v.as_ref()))
            .map(|(origin, target)| A::from(Multi { origin, target, identity: key }));
        Ok(arrow)
    }
}

impl Transaction<'_> {
    /// Check whether there exists any arrow of type `A` that points from `origin` to `target`.
    ///
    /// This only tells you whether there is *any* such arrow, not how many (in the case of parallel edges).
    pub fn exists<A>(&self, origin: Key, target: Key) -> Result<bool>
    where
        A: Arrow,
    {
        op::exists::<A>(self, origin, target)
    }
    /// Get all arrows of type `A` that point at `target`.
    pub fn incoming<'a, A>(&'a self, target: Key) -> impl Iterator<Item = Result<A>> + 'a
    where
        A: Arrow + 'a,
    {
        op::incoming::<A>(self, target).map_ok(A::from)
    }
    /// Get all arrows of type `A` that point away from `origin`.
    pub fn outgoing<'a, A>(&'a self, origin: Key) -> impl Iterator<Item = Result<A>> + 'a
    where
        A: Arrow + 'a,
    {
        op::outgoing::<A>(self, origin).map_ok(A::from)
    }
    /// Create a new arrow of type `A`.
    ///
    /// This operation supports both [`Multi`] and [`Basic`] arrows.
    ///
    /// # Example
    ///
    /// The following snippet creates an arrow between `origin` and `target`.
    ///
    /// ```rust
    /// # fn main () -> store::Result<()> {
    /// use store::{Arrow, Key};
    ///
    /// #[derive(Arrow)]
    /// struct MyArrow { origin: Key, target: Key }
    ///
    /// # let schema = store::types::Schema::new().has::<MyArrow>();
    /// # store::Store::test(schema, |db| {
    /// let origin = Key::gen();
    /// let target = Key::gen();
    ///
    /// db.run(|tx| {
    ///     tx.create(MyArrow { origin, target })
    /// })?;
    ///
    /// assert!(db.exists::<MyArrow>(origin, target)?);
    /// # store::OK })
    /// # }
    /// ```
    pub fn create<A>(&self, arrow: A) -> Result<()>
    where
        A: Arrow,
    {
        op::create::<A>(self, arrow.into())
    }
    /// Delete all edges of type `A` from `origin` to `target`.
    ///
    /// It is not an error for this function not to delete anything.
    pub fn delete_all<A>(&self, origin: Key, target: Key) -> Result<()>
    where
        A: Arrow,
    {
        op::delete_all::<A>(self, origin, target)
    }
    /// Delete a specific arrow.
    pub fn delete_one<A>(&self, arrow: A) -> Result<()>
    where
        A: Arrow,
    {
        op::delete_one::<A>(self, arrow.into())
    }
    /// List all arrows between `a` and `b`, in either direction.
    pub fn between<'a, A>(&'a self, a: Key, b: Key) -> impl Iterator<Item = Result<A>> + 'a
    where
        A: Arrow<Kind = Multi> + 'a,
    {
        op::between::<A>(self, a, b).map_ok(A::from)
    }
    /// Construct the arrow from its identifier.
    pub fn get_arrow<A>(&self, key: Key) -> Result<Option<A>>
    where
        A: Arrow<Kind = Multi>,
    {
        let arrow = self
            .open(crate::types::MULTIEDGE_HEADERS)
            .get(key)?
            .map(|v| Key::split(v.as_ref()))
            .map(|(origin, target)| A::from(Multi { origin, target, identity: key }));
        Ok(arrow)
    }
}

impl Batch {
    /// Create an arrow. See [`Transaction::create`].
    pub fn create<A>(&mut self, arrow: A)
    where
        A: Arrow,
    {
        op::create::<A>(self, arrow.into())
            .expect("no errors expected to occur during batch operation")
    }
    /// Delete a specific arrow.
    pub fn delete_one<A>(&mut self, arrow: A)
    where
        A: Arrow,
    {
        op::delete_one::<A>(self, arrow.into())
            .expect("no errors expected to occur during batch operation")
    }
}

mod op {
    //! Implementations of arrow operations.

    use super::*;
    use crate::{internal::*, types::MULTIEDGE_HEADERS, Key, Result, OK};

    /// Check whether there exists at least one arrow of type `A` from `origin` to `target`.
    pub fn exists<A>(cx: &impl Query, origin: Key, target: Key) -> Result<bool>
    where
        A: Arrow,
    {
        if A::Kind::IS_MULTI {
            // In the case of a multi-edge, at least one result from the prefix scan
            // indicates that there is at least one edge.
            cx.open(A::SPEC.by_origin)
                .scan(origin.fuse(target))
                .next()
                .transpose()
                .map(|o| o.is_some())
        } else {
            cx.open(A::SPEC.by_origin).has(origin.fuse(target))
        }
    }

    /// List incoming arrows relative to `target`.
    pub fn incoming<'db, A>(
        cx: &'db impl Query,
        target: Key,
    ) -> impl Iterator<Item = Result<A::Kind>> + 'db
    where
        A: Arrow,
        A::Kind: 'db,
    {
        // In the `by_target` keyspace, for either kind of arrow the layout is such that the target is
        // the prefix, so we pick that keyspace to more efficiently list all arrows that target the key.
        cx.open(A::SPEC.by_target)
            .scan(target)
            .map_ok(|(mut k, _)| {
                // Arrows from `by_target` are oriented target-first, while the decoder function requires
                // that the buffer is oriented origin-first. Regardless of whether `..32` covers the prefix
                // or the whole slice, swapping the two keys always gives us the ordering expected by the
                // decoding function.
                let (t, o) = k[..32].split_at_mut(16);
                t.swap_with_slice(o);
                A::Kind::dec(&k)
            })
    }

    /// List outgoing arrows relative to `origin`.
    pub fn outgoing<'db, A>(
        cx: &'db impl Query,
        origin: Key,
    ) -> impl Iterator<Item = Result<A::Kind>> + 'db
    where
        A: Arrow,
        A::Kind: 'db,
    {
        cx.open(A::SPEC.by_origin)
            .scan(origin)
            .map_ok(|(ref k, _)| A::Kind::dec(k))
    }

    /// Get all arrows between the two endpoints (in either direction)
    pub fn between<'db, A>(
        cx: &'db impl Query,
        origin: Key,
        target: Key,
    ) -> impl Iterator<Item = Result<A::Kind>> + 'db
    where
        A: Arrow,
        A::Kind: 'db,
    {
        let ks = cx.open(A::SPEC.by_origin);
        ks.scan(origin.fuse(target))
            .chain(ks.scan(target.fuse(origin)))
            .map_ok(|(ref k, _)| A::Kind::dec(k))
    }

    /// Create a new arrow.
    pub fn create<A>(cx: &impl Write, arrow: A::Kind) -> Result<()>
    where
        A: Arrow,
    {
        if A::Kind::IS_MULTI {
            let Multi { identity, origin, target } = unsafe { arrow.raw().multi };
            cx.open(MULTIEDGE_HEADERS)
                .set(identity, origin.fuse(target))?;
        }
        let (by_origin, by_target) = arrow.enc();
        cx.open(A::SPEC.by_origin).set(by_origin, b"")?;
        cx.open(A::SPEC.by_target).set(by_target, b"")?;
        OK
    }

    /// Delete all arrows from `origin` to `target`.
    ///
    /// TODO: Remove the query requirement (depends on range delete being available).
    pub fn delete_all<A>(cx: &(impl Write + Query), origin: Key, target: Key) -> Result<()>
    where
        A: Arrow,
    {
        let by_origin = cx.open(A::SPEC.by_origin);
        let by_target = cx.open(A::SPEC.by_target);
        Ok(if A::Kind::IS_MULTI {
            let headers = cx.open(MULTIEDGE_HEADERS);
            // TODO: optimize this implementation using range deletes.
            // Unfortunately, range deletes are not available in transactional backends.
            for key in by_origin.scan(origin.fuse(target)).keys() {
                let key = Multi::decode(key?.as_ref());
                by_origin.del(key.encode())?;
                by_target.del(key.swap().encode())?;
                headers.del(key.identity)?;
            }
        } else {
            by_origin.del(origin.fuse(target))?;
            by_target.del(target.fuse(origin))?;
        })
    }

    /// Delete a specific arrow, if it exists. Doesn't error if the arrow does *not* exist.
    pub fn delete_one<A>(cx: &impl Write, arrow: A::Kind) -> Result<()>
    where
        A: Arrow,
    {
        let (by_origin, by_target) = arrow.enc();
        cx.open(A::SPEC.by_origin).del(by_origin)?;
        cx.open(A::SPEC.by_target).del(by_target)?;
        OK
    }
}

/// Types representing the different kinds of arrows.
mod kinds {
    use super::ArrowKind;
    use crate::Key;

    impl ArrowKind for Multi {
        const IS_MULTI: bool = true;
        fn dec(buf: &[u8]) -> Self {
            Multi::decode(buf)
        }
        fn enc(self) -> (impl AsRef<[u8]>, impl AsRef<[u8]>) {
            (self.encode(), self.swap().encode())
        }
        fn raw(&self) -> super::Raw {
            super::Raw { multi: *self }
        }
    }

    impl ArrowKind for Basic {
        const IS_MULTI: bool = false;
        fn dec(buf: &[u8]) -> Self {
            Basic::decode(buf)
        }
        fn enc(self) -> (impl AsRef<[u8]>, impl AsRef<[u8]>) {
            (self.encode(), self.reverse().encode())
        }
        fn raw(&self) -> super::Raw {
            super::Raw { basic: *self }
        }
    }

    /// A multi-edge is an edge with an identity, which allows multiple parallel edges to exist
    /// between two vertices.
    #[derive(Clone, Copy)]
    pub struct Multi {
        /// The node that this edge points away from.
        pub origin: Key,
        /// The node that this edge points towards.
        pub target: Key,
        /// The discriminator of this particular edge, which distinguishes it from all other edges that
        /// connect `origin` and `target`, and indeed from every other edge or node in the graph.
        pub identity: Key,
    }

    impl Multi {
        /// Decode a multiarrow key from an origin-first encoded buffer. If the buffer is not correctly
        /// oriented, the results will be wrong; the arrow will be oriented *away* from the target and
        /// *at* the origin, instead of the other way around.
        ///
        /// # Orientation
        ///
        /// In this context, *correctly oriented* means that it is laid out in *origin-first* order,
        /// like this (where `o`, `t` and `i` represent bytes):
        ///
        /// ```text
        /// oooooooooooooooo tttttttttttttttt iiiiiiiiiiiiiiii
        /// |--------------| |--------------| |--------------|
        ///      origin           target          identity
        ///       ..16            16..32            32..
        /// ```
        ///
        /// In a *reverse oriented* buffer, the origin and target parts are swapped, so the target is
        /// the prefix, followed by the origin, and then the identity. This is also called *target-first*
        /// encoding in this documentation.
        ///
        /// # Silent failure
        ///
        /// There is no way to detect whether the ordering is correct from just the buffer, so the caller
        /// must ensure that the order is correct. If you have a target-first encoded buffer, you can have
        /// to swap the two keys before passing it into this function, or this function will give you an
        /// edge that does not exist (since a multiedge can only point in one direction).
        ///
        /// Safety-wise, this isn't an issue, so it does not warrant marking this function as `unsafe`.
        ///
        /// # Panics
        ///
        /// This function panics if `buf` is not exactly 48 bytes long.
        pub fn decode(buf: &[u8]) -> Multi {
            Multi {
                origin: Key::from_slice(&buf[..16]),
                target: Key::from_slice(&buf[16..32]),
                identity: Key::from_slice(&buf[32..]),
            }
        }
        /// Encode an arrow in *origin-first order*. See the docs of [`Multi::decode`] for an explanation
        /// of the difference between origin-first encoding and target-first encoding.
        pub fn encode(self) -> [u8; 48] {
            let mut key = [0; 48];
            key[..16].copy_from_slice(&self.origin.0);
            key[16..32].copy_from_slice(&self.target.0);
            key[32..].copy_from_slice(&self.identity.0);
            key
        }
        /// Swap the origin and target of this arrow, while leaving the identity the same.
        pub(super) fn swap(self) -> Multi {
            Multi {
                origin: self.target,
                target: self.origin,
                ..self
            }
        }
    }

    /// A normal directed edge. Duplicates are not allowed.
    ///
    /// This kind of arrow is useful for modeling predicates and simple relationships.
    #[derive(Clone, Copy)]
    pub struct Basic {
        pub origin: Key,
        pub target: Key,
    }

    impl Basic {
        /// Get the inverse of this arrow (an arrow that connects the same two nodes, but pointing in the
        /// other direction).
        pub fn reverse(self) -> Basic {
            Basic {
                origin: self.target,
                target: self.origin,
            }
        }
        /// Encode `self` in origin-first order. See [`Multi::decode`] for docs on ordering.
        pub fn encode(self) -> [u8; 32] {
            self.origin.fuse(self.target)
        }
        /// Decode a basic edge from a buffer laid out origin-first. See [`Multi::decode`] for more information
        /// about key encoding.
        ///
        /// # Panics
        ///
        /// Panics if `buf` is not exactly 32 bytes long.
        pub fn decode(buf: &[u8]) -> Basic {
            let (origin, target) = Key::split(buf);
            Basic { origin, target }
        }
    }
}
/// Derive [`Arrow`] for a struct.
///
/// This will generate the required [`Into`] and [`From`] impls, as well as an [`Arrow`](trait@Arrow) impl and
/// a [`DataType`] impl with the namespaces derived from the name of the struct. The macro works on structs with
/// specific fields, or newtypes of any arrow kind.
///
/// # Attributes
///
/// The `origin`, `target` and `identity` attributes are used on fields of type [`Key`], and they are used
/// to map the arrow's type to an [`ArrowKind`]. The `#[origin]` annotation isn't needed if the struct contains
/// a field named `origin`. Ditto with `target` and `identity`.
///
/// If there is no `identity` defined, the `ArrowKind` will be [`Basic`]. If an `identity` is defined, the kind
/// will be [`Multi`].
///
/// # Examples
///
/// Generates a [`Basic`] arrow called `my-arrow`.
///
/// ```
/// use store::{Key, Arrow, types::Schema};
///
/// #[derive(Arrow)]
/// struct MyArrow { origin: Key, target: Key }
///
/// // This will fail to compile if the type doesn't implement `Arrow` correctly
/// Schema::new().has::<MyArrow>();
/// ```
///
/// Newtypes of either arrow kind are supported.
///
/// ```
/// use store::{Key, arrow::{Basic, Multi, Arrow}};
///
/// /// The origin has requested to follow the target.
/// ///
/// /// Note: there may be more than one follow request between any two actors.
/// #[derive(Arrow)]
/// struct FollowRequest(Multi);
///
/// /// A relation between two actors meaning that the origin follows the target.
/// #[derive(Arrow)]
/// struct Follows(Basic);
///
/// /// Users can follow each other.
/// struct User(Key);
///
/// impl User {
///     /// Make `self` follow `other`.
///     pub fn follows(self, other: User) -> Follows {
///         Follows(Basic { origin: self.0, target: other.0 })
///     }
/// }
/// ```
///
/// Generates a [`Multi`] arrow called `my-multi-arrow`, mapping the multiarrow's discriminator to the struct's
/// `unique` field.
///
/// ```
/// use store::{Key, Arrow};
///
/// #[derive(Arrow)]
/// struct MyMultiArrow {
///     pub origin: Key,
///     pub target: Key,
///     #[identity]
///     pub unique: Key,
/// }
/// ```
///
/// The macro automatically adds `From` and `Into` implementations:
///
/// ```
/// use store::{Key, Arrow, arrow::Basic};
///
/// #[derive(Arrow)]
/// struct MyArrow { origin: Key, target: Key }
///
/// let origin = Key::gen();
/// let target = Key::gen();
///
/// let edge: Basic = MyArrow { origin, target }.into();
///
/// assert_eq!(origin, edge.origin);
/// assert_eq!(target, edge.target);
/// ```
pub use r#macro::Arrow;
