//! Defining a [`Schema`].
//!
//! There is a lot of complicated machinery here to make it so that you have to write very little code to
//! define new types. Basically, if you want to define a thing to store, you need to implement the trait
//! for it (e.g. [`Arrow`]), and also implement [`DataType`], where you create a specification describing which
//! namespaces store records of that type.
//!
//! Then, when you construct a new `Store`, you need to pass in a [`Schema`], or the database won't be able
//! to operate on the types.
//!
//! [`Arrow`]: super::Arrow

use std::collections::HashSet;

use derive_more::Display;

/// The namespace where all vertices must be registered.
pub(crate) const NODE_HEADERS: Keyspace = Keyspace("header:node");

/// The namespace where multiedge identities are mapped to endpoints.
pub(crate) const MULTIEDGE_HEADERS: Keyspace = Keyspace("header:multiedge");

/// A specification of all user-defined namespaces.
///
/// # Example
///
/// The below example correctly defines a [basic arrow] and demonstrates its use by inserting one and then
/// testing whether it exists. If the appropriate keyspaces are not known to the store, this will panic.
///
/// ```rust
/// use store::{ arrow::Arrow, types::Schema, Store, Key, OK };
///
/// // Each kind of value has a derive macro.
/// #[derive(Arrow)]
/// struct MyArrow { origin: Key, target: Key }
///
/// fn main () -> store::Result<()> {
///     // Here, we make sure that the namespaces used for `MyArrow` are known.
///     let schema = Schema::new()
///         .has::<MyArrow>();
///
///     let result = Store::test(schema, |db| {
///         let origin = Key::gen();
///         let target = Key::gen();
///
///         let mut changes = db.batch();
///         changes.create(MyArrow { origin, target });
///         db.apply(changes)?;
///
///         db.exists::<MyArrow>(origin, target)
///     })?;
///
///     assert!(result);
///     OK
/// }
/// ```
///
/// [basic arrow]: crate::arrow::Basic
pub struct Schema(pub(crate) HashSet<Keyspace>);

impl Schema {
    /// Construct a new empty schema.
    pub fn new() -> Schema {
        Schema(HashSet::from_iter([NODE_HEADERS, MULTIEDGE_HEADERS]))
    }
    /// Add the component to the schema.
    pub fn has<C>(mut self) -> Schema
    where
        C: DataType,
    {
        self.add(C::SPEC);
        self
    }
    /// Add a spec to the schema by mutable reference.
    pub fn add(&mut self, spec: impl TypeSpec) -> &mut Schema {
        spec.register(&mut self.0);
        self
    }
}

/// The name of a keyspace.
///
/// Specifically, this is the name of a RocksDB column family.
#[derive(Clone, Copy, PartialEq, Eq, Hash, PartialOrd, Ord, Display)]
pub struct Keyspace(pub &'static str);

impl AsRef<str> for Keyspace {
    fn as_ref(&self) -> &str {
        self.0
    }
}

/// A *thing* that is stored in the database, either an [arrow](ArrowSpec), an [alias](AliasSpec), or a
/// [mixin](MixinSpec).
///
/// All namespaces must be unique, and added to the [`Schema`].
pub trait DataType {
    type Type: TypeSpec;
    const SPEC: Self::Type;
}

/// The specification for an [`Arrow`](crate::Arrow).
///
/// The listed namespaces must be unique among all other namespaces.
#[derive(Clone, Copy)]
pub struct ArrowSpec {
    /// The keyspace where edge keys are ordered `(origin, target)`.
    pub by_origin: Keyspace,
    /// The keyspace where edge keys are ordered `(target, origin)`.
    pub by_target: Keyspace,
}

/// A specification for the namespaces needed to store an [`Alias`][crate::Alias].
#[derive(Clone, Copy)]
pub struct AliasSpec {
    /// The alias -> key mapping table.
    pub keyspace: Keyspace,
    /// The key -> alias mapping table.
    pub reversed: Keyspace,
}

/// Where do we store a mixin?
#[derive(Clone, Copy)]
pub struct MixinSpec {
    /// The key -> mixin mapping table.
    pub keyspace: Keyspace,
}

/// Describes how to add a [`DataType`] to a [`Schema`].
pub trait TypeSpec {
    /// Register the namespaces.
    fn register(&self, set: &mut HashSet<Keyspace>);
}

// TODO: better error messages.

impl TypeSpec for ArrowSpec {
    fn register(&self, set: &mut HashSet<Keyspace>) {
        if !set.insert(self.by_origin) {
            panic! {
                "Duplicate found while inserting Arrow::BY_ORIGIN: {}",
                self.by_origin
            }
        }
        if !set.insert(self.by_target) {
            panic! {
                "Duplicate found while inserting Arrow::BY_TARGET: {}",
                self.by_target
            }
        }
    }
}
impl TypeSpec for AliasSpec {
    fn register(&self, set: &mut HashSet<Keyspace>) {
        if !set.insert(self.keyspace) {
            panic! {
                "Duplicate found while inserting Alias::KEYSPACE: {}",
                self.keyspace
            }
        }
        if !set.insert(self.reversed) {
            panic! {
                "Duplicate found while inserting Alias::REVERSED: {}",
                self.reversed
            }
        }
    }
}
impl TypeSpec for MixinSpec {
    fn register(&self, set: &mut HashSet<Keyspace>) {
        if !set.insert(self.keyspace) {
            panic! {
                "Duplicate found while inserting Mixin::KEYSPACE: {}",
                self.keyspace
            }
        }
    }
}
